/*----------------------------------------------------------------------*\
 * xcolrsel.c  (Select and display colors from the X11 rgb.txt file)	*
 * Version 1.1								*
 * Based on bits and pieces of xact, xrgb, and xfontsel.		*
 * /UCAR/NCAR/ATD/SSSF/RDP/ G. Forrest Cook  4/26/94			*
\*----------------------------------------------------------------------*/

/*

	(C) Copyright 1994
	University Corporation for Atmospheric Research

The X Consortium, and any party obtaining a copy of these files from
the X Consortium, directly or indirectly, is granted, free of charge, a
full and unrestricted irrevocable, world-wide, paid up, royalty-free,
nonexclusive right and license to deal in this software and
documentation files (the "Software"), including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons who receive
copies from any such party to do so.  This license includes without
limitation a license to do the foregoing actions under any patents of
the party supplying this software to the X Consortium.

The above copyright notice and this permission notice shall be included
in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
IN NO EVENT SHALL THE X CONSORTIUM BE LIABLE FOR ANY CLAIM, DAMAGES OR
OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
OTHER DEALINGS IN THE SOFTWARE.

*/

/*
 * Version History
 * 1.0  First version
 * 1.1  The code was cleaned up and unnecessary globals were eliminated.
 */

#include <stdio.h>
#include <fcntl.h>
#include <X11/Xatom.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Shell.h>
#include <X11/Xaw/AsciiText.h>
#include <X11/Xaw/Form.h>
#include <X11/Xaw/Label.h>
#include <X11/Xaw/Command.h>
#include <X11/Xaw/Cardinals.h>

#define XBG 0		/* constant for background */
#define XFG 1		/* constant for foreground */
#define FNSTLEN 120	/* max filename length */
#define MAXXCOLS 1000	/* Max number of colors in rgb.txt */
#define MAXXCNLN 32	/* Max color name length for rgb.txt entries */

void exit_prog();
void l_select(), r_select(), lff_sel(), rff_sel();
void lf8_sel(), rt8_sel(), set_bgfg();
void col_select(), set_bg_color(), set_fg_color();
Boolean ConvertSelection();

/*
 * Widgets that need to be global
 */
Widget Top, Bgfg, Bcname, Fcname;

/*
 * Other global variables
 */
char *Col_string;
int Nxcols = 0;		/* actual count of used colors from rgb.txt */
int Xbgpos = 0;		/* background color position */
int Xfgpos = 0;		/* foreground color position */
int Xbgfg = XBG;	/* flag for active background or foreground */
XColor Xc[2];		/* one color for background, one for foreground */

/*
 * Here's where the rgb.txt info gets stored as RGB primary levels and a name.
 */
int ColR[MAXXCOLS];
int ColG[MAXXCOLS];
int ColB[MAXXCOLS];
char ColNm[MAXXCOLS][MAXXCNLN];

/*----------------------------------------------------------------------*/
/*
 * Define a bunch of fallback resources.
 */

static String fallback_resources[] =
{
	"*title:			xcolrsel  1.1",
	"*background:			black",
	"*foreground:			white",
	"*Form.background:		gray",
	"*Label.foreground:		yellow",
	"*Label.resize:			FALSE",
	"*Label.width:			130",
	"*Label.height:			24",
	"*form.defaultDistance:		2",
	"*bgfg.label:			bg",
	"*bgfg.foreground:		orange",
	"*lff.label:			<<",
	"*lff.foreground:		green",
	"*lf8.label:			<8",
	"*lf8.foreground:		green",
	"*left.label:			<",
	"*left.foreground:		green",
	"*right.label:			>",
	"*right.foreground:		green",
	"*rt8.label:			8>",
	"*rt8.foreground:		green",
	"*rff.label:			>>",
	"*rff.foreground:		green",
	"*bcname.foreground:		yellow",
	"*bcname.width:			211",
	"*fcname.foreground:		yellow",
	"*fcname.width:			211",
	"*fpform.transient: 		False",
	"*quit.width:			43",
	"*quit.label:			Quit",
	"*quit.foreground:		red",
	"*mix.height:			80",
	"*mix.width:			211",
	"*mix.label:			Xcolrsel Test Text",
	NULL
};

/*----------------------------------------------------------------------*/

main (argc, argv)
int argc;
char **argv;
{
	int i, n;
	int fgf=XFG, bgf=XBG;
	char selbuf[120], tmpstr[80];
	char filename[FNSTLEN];
	FILE *cf;
	Arg args[10];
	Widget w, form, lqu, lff, lf8, left, right, rt8, rff;
	Widget cbar, h=NULL, v=NULL;
	XtAppContext actx;
	XtTranslations seltrans;

        get_opt_str_arg (argc, argv, 'f', RGBFILE, filename);

/*
 * Open the rgb.txt file and read in some colors.
 */
	if ((cf = fopen (filename, "r")) == NULL)
	{
	printf ("\nError: xcolrsel couldn't open file: %s\n", filename);
		exit (1);
	}

	i=0;

	do
	{
		fgets (tmpstr, 80, cf);

		/*
		 * Lines starting with ! are comments, skip them.
		 */
		if (tmpstr[0] != '!')
		{
		    sscanf (tmpstr, "%d %d %d %s",
			&ColR[i], &ColG[i], &ColB[i], ColNm[i]);

	/*printf ("%d %d %d %s\n", ColR[i], ColG[i], ColB[i], ColNm[i]);*/

		    /*
		     * Detect duplicate colors and keep the last one and
		     * blindly assume that it has a name without whitespace.
		     */
		    if (i > 0)
		    {
			if ((ColR[i] == ColR[i-1]) &&
			    (ColG[i] == ColG[i-1]) &&
			    (ColB[i] == ColB[i-1]))
			{
				strcpy (ColNm[i-1], ColNm[i]);
				i--;
			}
		    }

		    i++;
		}
	}
	while (!feof(cf));

	fclose (cf);

	Nxcols = i;

#ifdef DEBUG
	for (i=0; i<Nxcols; i++)
		printf ("%d %d %d %s\n", ColR[i], ColG[i], ColB[i], ColNm[i]);
#endif
/*
 * Set the starting background and foreground color values.
 */
	for (i=0; i<=1; i++)
	{
		Xc[i].flags = DoRed | DoGreen | DoBlue;
	}
/*
 * Create an application shell, and a form to go inside it.
 */
	Top = XtAppInitialize (&actx, "Xcolrsel", (XrmOptionDescList)NULL,ZERO,
			&argc, argv, fallback_resources, (ArgList)NULL, ZERO);

	form = XtCreateManagedWidget ("form",formWidgetClass,Top,NULL,ZERO);
/*
 * Create a quit button.
 */
	n = 0;
	XtSetArg (args[n], XtNfromHoriz, h);		n++;
	XtSetArg (args[n], XtNfromVert, v);		n++;
	h=lqu=XtCreateManagedWidget ("quit", commandWidgetClass,form,args,n);
	XtAddCallback (lqu, XtNcallback, exit_prog, (void *)0);
/*
 * Create a left fast forward color select button.
 */
	n = 0;
	XtSetArg (args[n], XtNfromHoriz, h);		n++;
	XtSetArg (args[n], XtNfromVert, v);		n++;
	h=lff=XtCreateManagedWidget ("lff", commandWidgetClass, form,args,n);
	XtAddCallback (lff, XtNcallback, lff_sel, (void *)0);
/*
 * Create a left 8 color select button.
 */
	n = 0;
	XtSetArg (args[n], XtNfromHoriz, h);		n++;
	XtSetArg (args[n], XtNfromVert, v);		n++;
	h=lf8=XtCreateManagedWidget ("lf8", commandWidgetClass, form,args,n);
	XtAddCallback (lf8, XtNcallback, lf8_sel, (void *)0);
/*
 * Create a move left color select button.
 */
	n = 0;
	XtSetArg (args[n], XtNfromHoriz, h);		n++;
	XtSetArg (args[n], XtNfromVert, v);		n++;
	h=left=XtCreateManagedWidget ("left",commandWidgetClass, form,args,n);
	XtAddCallback (left, XtNcallback, l_select, (void *)0);
/*
 * Create a move right color select button.
 */
	n = 0;
	XtSetArg (args[n], XtNfromHoriz, h);		n++;
	XtSetArg (args[n], XtNfromVert, v);		n++;
	h=right=XtCreateManagedWidget("right",commandWidgetClass,form,args,n);
	XtAddCallback (right, XtNcallback, r_select, (void *)0);
/*
 * Create a right 8 color select button.
 */
	n = 0;
	XtSetArg (args[n], XtNfromHoriz, h);		n++;
	XtSetArg (args[n], XtNfromVert, v);		n++;
	h=rt8=XtCreateManagedWidget ("rt8", commandWidgetClass, form,args,n);
	XtAddCallback (rt8, XtNcallback, rt8_sel, (void *)0);
/*
 * Create a right fast forward color select button.
 */
	n = 0;
	XtSetArg (args[n], XtNfromHoriz, h);		n++;
	XtSetArg (args[n], XtNfromVert, v);		n++;
	h=rff=XtCreateManagedWidget ("rff", commandWidgetClass,form,args,n);
	XtAddCallback (rff, XtNcallback, rff_sel, (void *)0);
/*
 * Create a background/foreground button.
 */
        n = 0;
        XtSetArg (args[n], XtNfromHoriz, h);            n++;
        XtSetArg (args[n], XtNfromVert, v);             n++;
        h=Bgfg=XtCreateManagedWidget ("bgfg", commandWidgetClass,form,args,n);
        XtAddCallback (Bgfg, XtNcallback, set_bgfg, (void *)0);
/*
 * Row 2
 */
	h=NULL;
	v=lff;
/*
 * Create a background color name widget.
 */
	n = 0;
	XtSetArg (args[n], XtNfromHoriz, h);		n++;
	XtSetArg (args[n], XtNfromVert, v);		n++;
	h=Bcname=XtCreateManagedWidget ("bcname",commandWidgetClass,
					form,args,n);
	XtAddCallback (Bcname, XtNcallback, col_select, &bgf);
/*
 * Row 3
 */
	h=NULL;
	v=Bcname;
/*
 * Create a foreground color name widget.
 */
	n = 0;
	XtSetArg (args[n], XtNfromHoriz, h);		n++;
	XtSetArg (args[n], XtNfromVert, v);		n++;
	h=Fcname=XtCreateManagedWidget ("fcname",commandWidgetClass,
					form,args,n);
	XtAddCallback (Fcname, XtNcallback, col_select, &fgf);
/*
 * Row 4
 */
	h=NULL;
	v=Fcname;
/*
 * Get one color cell for the foreground and one for the background.
 */
	for (i=0; i<=1; i++)
	{
		XAllocColorCells (XtDisplay (Top),
			DefaultColormap(XtDisplay(Top), 0),
			False, NULL, 0, &Xc[i].pixel, 1);
	}

/*
 * Create a label widget to show the color.
 */
	n = 0;
	XtSetArg (args[n], XtNfromHoriz, h);		n++;
	XtSetArg (args[n], XtNfromVert, v);		n++;
	XtSetArg (args[n], XtNbackground, Xc[0].pixel); n++;
	XtSetArg (args[n], XtNforeground, Xc[1].pixel); n++;
	cbar = XtCreateManagedWidget ("mix", labelWidgetClass, form,args,n);

/*
 * Display the top widget.
 */
	XtRealizeWidget (Top);

/*
 * Set up the first color.
 */
	set_bg_color ();
	set_fg_color ();

/*
 * Here's where it all happens.
 */
	XtAppMainLoop (actx);
}

/*----------------------------------------------------------------------*/
/*
 * Set the color label background to the current color and set the name string.
 */
void set_bg_color ()
{
	Xc[0].red = ColR[Xbgpos] << 8;
	Xc[0].green = ColG[Xbgpos] << 8;
	Xc[0].blue = ColB[Xbgpos] << 8;
	XStoreColor(XtDisplay(Top),DefaultColormap(XtDisplay(Top),0),&(Xc[0]));

	SetLabelStr (Bcname, ColNm[Xbgpos]);
}

/*----------------------------------------------------------------------*/
/*
 * Set the color label foreground to the current color and set the name string.
 */
void set_fg_color ()
{
	Xc[1].red = ColR[Xfgpos] << 8;
	Xc[1].green = ColG[Xfgpos] << 8;
	Xc[1].blue = ColB[Xfgpos] << 8;
	XStoreColor(XtDisplay(Top),DefaultColormap(XtDisplay(Top),0),&(Xc[1]));

	SetLabelStr (Fcname, ColNm[Xfgpos]);
}

/*----------------------------------------------------------------------*/
/*
 * Exit the program here.
 */
void exit_prog ()
{
	exit (0);
}

/*----------------------------------------------------------------------*/
/*
 * Set background or foreground mode
 */
void set_bgfg ()
{
	Arg newargs[1];

	if (Xbgfg == XBG)
	{
		Xbgfg = XFG;
		XtSetArg (newargs[0], "label", "fg");
	}
	else
	{
		Xbgfg = XBG;
		XtSetArg (newargs[0], "label", "bg");
	}
        XtSetValues (Bgfg, newargs, 1);
}

/*----------------------------------------------------------------------*/
/*
 * Move to the beginning color
 */
void lff_sel ()
{
	if (Xbgfg == XBG)
	{
		Xbgpos = 0;
		set_bg_color ();
	}
	else
	{
		Xfgpos = 0;
		set_fg_color ();
	}
}

/*----------------------------------------------------------------------*/
/*
 * Move back 8 colors.
 */
void lf8_sel ()
{
	if (Xbgfg == XBG)
	{
		Xbgpos -= 8;
		if (Xbgpos < 0) Xbgpos = 0;
		set_bg_color ();
	}
	else
	{
		Xfgpos -= 8;
		if (Xfgpos < 0) Xfgpos = 0;
		set_fg_color ();
	}
}

/*----------------------------------------------------------------------*/
/*
 * Move back one color
 */
void l_select (w, closure, callData)
Widget w;
XtPointer closure, callData;
{
	if (Xbgfg == XBG)
	{
		if (Xbgpos > 0) Xbgpos--;
		set_bg_color ();
	}
	else
	{
		if (Xfgpos > 0) Xfgpos--;
		set_fg_color ();
	}
}

/*----------------------------------------------------------------------*/
/*
 * Move up one color
 */
void r_select ()
{
	if (Xbgfg == XBG)
	{
		if (Xbgpos < (Nxcols-1)) Xbgpos++;
		set_bg_color ();
	}
	else
	{
		if (Xfgpos < (Nxcols-1)) Xfgpos++;
		set_fg_color ();
	}
}

/*----------------------------------------------------------------------*/
/*
 * Move up 8 colors.
 */
void rt8_sel ()
{
	if (Xbgfg == XBG)
	{
		Xbgpos += 8;
		if (Xbgpos >= Nxcols) Xbgpos = Nxcols-1;
		set_bg_color ();
	}
	else
	{
		Xfgpos += 8;
		if (Xfgpos >= Nxcols) Xfgpos = Nxcols-1;
		set_fg_color ();
	}
}

/*----------------------------------------------------------------------*/
/*
 * Move to the end color
 */
void rff_sel ()
{
	if (Xbgfg == XBG)
	{
		Xbgpos = Nxcols-1;
		set_bg_color ();
	}
	else
	{
		Xfgpos = Nxcols-1;
		set_fg_color ();
	}
}

/*----------------------------------------------------------------------*/

SetLabelStr (wi, str)  /* put a string into a label widget */
Widget wi;
char *str;
{
        int n = 0;
        Arg args[1];

        XtSetArg (args[n], XtNlabel, str);             n++;
        XtSetValues (wi, args, n);
}
/*----------------------------------------------------------------------*\
 * The following function was grabbed from xfontsel and modified.	*
 * 									*
 * Copyright 1985, 1986, 1987, 1988, 1989 by the			*
 * Massachusetts Institute of Technology				*
\*----------------------------------------------------------------------*/

Boolean
ConvertSelection(w, selection, target, type, value, length, format)
    Widget w;
    Atom *selection, *target, *type;
    XtPointer *value;
    unsigned long *length;
    int *format;
{
	/* XmuConvertStandardSelection will use the second parameter only when
	 * converting to the target TIMESTAMP.  However, it will never be
	 * called upon to perform this conversion, because Xt will handle it
	 * internally.  CurrentTime will never be used.
	 */

	if (XmuConvertStandardSelection(w, CurrentTime, selection, target,
		type, (caddr_t *) value, length, format))
		return True;

	if (*target == XA_STRING)
	{
		*type = XA_STRING;
		*value = Col_string;
		*length = strlen(*value);
		*format = 8;
		return True;
	}
	else return False;
}

/*----------------------------------------------------------------------*/

void
col_select (w, fgbg)
Widget w;
int *fgbg;
{
	int pos;
	Time time = XtLastTimestampProcessed(XtDisplay(w));

/*
 * Set pos to index either the foreground or background color
 * name depending on which select button was pushed.
 */
	if (*fgbg == XBG)
		pos = Xbgpos;
	else	pos = Xfgpos;

/*
 * Grab some memory for the selection string, it gets freed externally.
 * Stuff a color name into the string.
 */
	Col_string = XtMalloc (sizeof (ColNm[pos]));

	strcpy (Col_string, ColNm[pos]);

	XtOwnSelection (w, XA_PRIMARY, time, ConvertSelection, NULL, NULL);
}



/*----------------------------------------------------------------------*\
 * Search the argv strings for an occurrence of a "-[char] string" arg,  *
 * if it is found, set retstr to the string value of the next argument. *
 * Return 1 for success, 0 for failure.                                 *
\*----------------------------------------------------------------------*/

int get_opt_str_arg (argc, argv, argchr, dflt, retstr)
int argc;
char **argv, argchr;
char *dflt, *retstr;
{
        int i=1, sl, sn;
        static char cmpstr[3] = "-?";
        cmpstr[1] = argchr;

        while (i < argc)
        {
                if (!strcmp(cmpstr, argv[i]))
                {
                        if (i < argc-1)
                        {
                                sl = strlen (argv[i+1]);
                                if (sl)
                                {
                                        strncpy (retstr, argv[i+1], sl);
                                        return (1);
                                }
                        }
                }
                i++;
        }

        strcpy (retstr, dflt);

        return (0);
}

/*----------------------------------------------------------------------*/
